# Copyright (c) Alibaba, Inc. and its affiliates.
from modelscope.hub import constants

constants.API_HTTP_CLIENT_TIMEOUT = 5
constants.API_FILE_DOWNLOAD_TIMEOUT = 300
constants.API_FILE_DOWNLOAD_CHUNK_SIZE = 1024 * 1024 * 16
